/*****************************************************************************
 *  Gap2Seq
 *  Copyright (C) Leena Salmela, Kristoffer Sahlin, Veli Mkinen,
 *  Alexandru Tomescu 2014
 *
 *  Contact: leena.salmela@cs.helsinki.fi
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*****************************************************************************/

#include <Gap2Seq.hpp>
#include <map>
#include <fstream>
#include <unordered_map>
#include <unordered_set>
using namespace std;

// Maximum number of paths for counting
#define MAX_PATHS (INT_MAX/2-1)

// Deafault maximum memory usage of DP table per thread
#define DEFAULT_MAX_MEMORY "20"

// Check if a file is readable
bool is_readable( const std::string & file ) 
{ 
    std::ifstream f( file.c_str() ); 
    return !f.fail(); 
} 


/********************************************************************************/

// Constants for command line parameters
static const char* STR_KMER_LEN = "-k";
static const char* STR_SOLID_THRESHOLD = "-solid";
static const char* STR_READS = "-reads";
static const char* STR_SCAFFOLDS = "-scaffolds";
static const char* STR_FILLED_SCAFFOLDS = "-filled";
static const char* STR_DIST_ERROR = "-dist-error";
static const char* STR_FUZ = "-fuz";
static const char* STR_MAX_MEM = "-max-mem";

/*********************************************************************
Constructor for the tool.
*********************************************************************/
Gap2Seq::Gap2Seq ()  : Tool ("Gap2Seq")
{
    // We add some custom arguments for command line interface
    getParser()->push_front (new OptionOneParam (STR_KMER_LEN, "kmer length",  false, "31"));
    getParser()->push_front (new OptionOneParam (STR_SOLID_THRESHOLD, "Threshold for solid k-mers",  false, "2"));
    getParser()->push_front (new OptionOneParam (STR_READS, "FASTA/Q files of reads. For several files use a comma separated list.",  true));
    getParser()->push_front (new OptionOneParam (STR_SCAFFOLDS, "FASTA/Q file of scaffolds",  true));
    getParser()->push_front (new OptionOneParam (STR_FILLED_SCAFFOLDS, "FASTA file of filled scaffolds",  true));
    getParser()->push_front (new OptionOneParam (STR_DIST_ERROR, "Maximum error in gap estimates",  false, "500"));
    getParser()->push_front (new OptionOneParam (STR_FUZ, "Number of nucleotides to ignore on gap fringes",  false, "10"));
    getParser()->push_front (new OptionOneParam (STR_MAX_MEM, "Maximum memory usage of DP table computation in gigabytes (excluding DBG)",  false, DEFAULT_MAX_MEMORY));
}

#ifndef SINGLE_THREAD
// Synchronizer for access to global variables
ISynchronizer *lock;
#endif

/*********************************************************************
Main method for gap filling
*********************************************************************/
void Gap2Seq::execute ()
{
  // Initialize random number generation for choosing random paths
  srand(time(NULL));

  // Get the command line arguments
  int k = getInput()->getInt(STR_KMER_LEN);
  int solid = getInput()->getInt(STR_SOLID_THRESHOLD);
  std::string reads = getInput()->getStr(STR_READS);
  std::string scaffolds = getInput()->getStr(STR_SCAFFOLDS);
  std::string filled_scaffolds = getInput()->getStr(STR_FILLED_SCAFFOLDS);
  int d_err = getInput()->getInt(STR_DIST_ERROR);
  int fuz = getInput()->getInt(STR_FUZ);
  long long max_mem = (long long)(getInput()->getDouble(STR_MAX_MEM) * 1024*1024*1024);
  std::string readsGraph = reads + ".h5";

  std::cout << "k-mer size: " << k << std::endl;
  std::cout << "Solidity threshold: " << solid << std::endl;
  std::cout << "Reads file: " << reads << std::endl;
  std::cout << "Scaffolds file: " << scaffolds << std::endl;
  std::cout << "Filled scaffolds file: " << filled_scaffolds << std::endl;
  std::cout << "Distance error: " << d_err << std::endl;
  std::cout << "Fuz: " << fuz << std::endl;
  std::cout << "Max memory: " << max_mem << std::endl;

  // Create de bruijn graph
  Graph graph;
  if (is_readable(readsGraph)) {
    std::cout << "Loading from " << readsGraph << std::endl;
    graph = Graph::load(reads);
  } else {
    // Tokenize reads file (list of files separated by ,)
    int filecount=1;
    char *readcstr = (char *)reads.c_str();
    for(int i = 0; i < strlen(readcstr); i++) {
      if (readcstr[i] == ',')
	filecount++;
    }
    
    if (filecount > 1) {
      char **files = new char*[filecount];
      files[0] = readcstr;
      int j = 1;
      int l = strlen(readcstr);
      for (int i = 0; i < l; i++) {
	if (readcstr[i] == ','){
	  readcstr[i] = '\0';
	  files[j] = &readcstr[i+1];
	  j++;
	}
      }
      try {
	BankFasta *b = new BankFasta(filecount, files);
	graph = Graph::create(b, (char const *)"-kmer-size %d -abundance %d -bloom cache -debloom original", k, solid);
      } catch (Exception& e) {
	std::cout << "DBG building failed: " << e.getMessage() << std::endl;
	exit(EXIT_FAILURE);
      }
    } else {
      try {
	graph = Graph::create ((char const *)"-in %s -kmer-size %d -abundance %d -bloom cache -debloom original", reads.c_str(), k, solid);
      } catch (Exception& e) {
	std::cout << "DBG building failed: " << e.getMessage() << std::endl;
	exit(EXIT_FAILURE);
      }
    }
  }
  std::cout << graph.getInfo() << std::endl;

  // Open the input scaffold file
  BankFasta scaffoldBank(scaffolds);

  // Iterate over scaffolds
  BankFasta::Iterator itSeq(scaffoldBank);

  // Open the output scaffold file
  BankFasta output(filled_scaffolds);

  // Count gaps and filled gaps
  int gapcount = 0;
  int filledgapcount = 0;

#ifndef SINGLE_THREAD
  // Synchronization for output (both file and stdout)
  lock = System::thread().newSynchronizer();
  IDispatcher *d = getDispatcher();
  // Default group size is too big (one thread will do all the work...)
  d->setGroupSize(1);

  max_mem = max_mem / d->getExecutionUnitsNumber();
#endif

  std::cout << "Max mem: " << max_mem << std::endl;

  // Parallel code starts here
#ifdef SINGLE_THREAD
  for(itSeq.first(); !itSeq.isDone();itSeq.next()) {
    // Original sequence
    std::string seq = itSeq->toString();
    // std::string seq = itSeq->toString();
#else
  // IDispatcher::Status status = d->iterate(&itSeq, [&] (const Sequence& sequence) {    
  Range<int>::Iterator intIt(1,d->getExecutionUnitsNumber());
  itSeq.first();
  IDispatcher::Status status = d->iterate(intIt, [&] (int id) {
    bool more = true;

    while (more) {
    std::string seq = "";
    std::string comment = "";
    {
      LocalSynchronizer local(lock);
      if (!itSeq.isDone()) {
	seq = itSeq->toString();
	comment = itSeq->getComment();
	itSeq.next();
	more = true;
      } else {
	more = false;
      }
    }
    if (!more)
      break;
#endif
      
    //    std::string seq = sequence.toString();


    // Gap filled sequence
    std::string filledSeq = "";
    // Position where previous gap ended. The filled sequence is
    // complete up to this position (in the original sequence)
    int prevGapEnd = 0;
    // Index over sequence positions
    int i=0;

#ifdef SINGLE_THREAD
    std::cout << itSeq->getComment() << std::endl;
#else
    {
      LocalSynchronizer local(lock);
      std::cout << comment << std::endl;

#ifdef DEBUG
      std::cout << seq << std::endl;
#endif
    }
#endif

    // Iterate over the sequence and find gaps
    while (i < seq.size()) {
      // A gap starts here
      if (seq[i] == 'N' || seq[i] == 'n') {
#ifndef SINGLE_THREAD
	{
	  LocalSynchronizer local(lock);
#endif
	  gapcount++;
#ifndef SINGLE_THREAD
	}
#endif
	// The first possible starting position for left k-mer
	int kmer_start = i-k-fuz;
	// The length of gap
	int gap = 0;
	// Measure the gap length
	while (i < seq.size() && (seq[i] == 'N' || seq[i] == 'n')) {
	  i++;
	  gap++;
	}

	// Check that the end k-mer is complete, i.e. no N's and enough sequence
	int ok = i+k+fuz <= seq.size() ? 1 : 0;
	for(int j = 0; j < k+fuz && ok; j++) {
	  if (seq[i+j] == 'N' || seq[i+j] == 'n')
	    ok = 0;
	}

	// Check that the start k-mer is complete
	if (kmer_start >= prevGapEnd && ok) {
	  // Number of paths found
	  int s;
	  // Buffer to hold the sequence to fill the gap
	  char *buf = new char[gap+k+d_err+2*fuz+1];
	  // How many k-mers on the edge of the gap are ignored by the paths
	  int left_fuz=0, right_fuz=0;

	  s = fill_gap(graph, seq.substr(kmer_start,k+fuz), seq.substr(i,k+fuz), gap+k, d_err, fuz, k, max_mem, &left_fuz, &right_fuz, buf);

#ifndef SINGLE_THREAD
	  {
	    LocalSynchronizer local(lock);
#endif

	    if (s > 0) {
	      filledgapcount++;
	      std::cout << "GapStart: " << kmer_start+k+fuz << " GapEnd: " << i << 
		" GapLength: " << gap << " PathsFound: " << s << " FilledGapLength: " << (int)strlen(&buf[fuz-left_fuz])-k << 
		" LeftFuz: " << left_fuz << " RightFuz: " << right_fuz << std::endl;
	    } else {
	      std::cout << "GapStart: " << kmer_start+k+fuz << " GapEnd: " << i <<
		" GapLength: " << gap << " PathsFound: " << s << " FilledGapLength: " << 0 << 
		" LeftFuz: " << left_fuz << " RightFuz: " << right_fuz << std::endl;
	    }
#ifndef SINGLE_THREAD
	  }
#endif
	  if (s > 0) {
	    // At least one path was found

#ifdef DEBUG
	    std::cout << "Fill: " << &buf[fuz-left_fuz] << std::endl;
#endif
	    // Add seq from end of previous gap to gap start and the gap fill sequence
	    filledSeq = filledSeq + seq.substr(prevGapEnd, kmer_start+k+fuz-left_fuz-prevGapEnd)+(string)&buf[fuz-left_fuz];
	    // Remove the right kmer
	    filledSeq = filledSeq.substr(0,filledSeq.length()-k);
	    // adjust gap end index
	    i += right_fuz;
	  } else {
	    // No paths found, keep the gap
	    // Add seq from end of previous gap to end of this gap
	    filledSeq = filledSeq + seq.substr(prevGapEnd, kmer_start+k+fuz+gap-prevGapEnd);
	  }
	  delete [] buf;
	} else {
	  // Either left or right k-mer is not complete
	  // Add seq from end of previous gap to end of this gap
	  filledSeq = filledSeq + seq.substr(prevGapEnd, kmer_start+k+fuz+gap-prevGapEnd);
	}
	// This gap has been processed
	prevGapEnd = i;
      } else {
	// No gap at position i
	i++;
      }
    }
    // Add the end of the seq
    filledSeq = filledSeq + seq.substr(prevGapEnd, seq.length()-prevGapEnd);

    // Write the filled scaffold to file
    Sequence s((char *)filledSeq.c_str());
#ifdef SINGLE_THREAD
    s._comment = itSeq->getComment();
#else
    s._comment = comment;
#endif
#ifndef SINGLE_THREAD
    {
      LocalSynchronizer local(lock);
#endif
      output.insert(s);
      output.flush();
#ifndef SINGLE_THREAD
    }
#endif
#ifdef DEBUG
    std::cout << filledSeq << std::endl;
#endif
#ifdef SINGLE_THREAD
    }
#else
    }
    });
#endif
  output.flush();

  std::cout << "Filled " << filledgapcount << " gaps out of " << gapcount << std::endl;

}


// Threadwise memory limit using std allocators

// Array of memory usage accessed by thread id
unordered_map<long long, long long> memuse;

// Custom memory allocator to be used with stl containers
template <typename T>
class count_allocator: public std::allocator<T> {
public:
  typedef size_t size_type;
  typedef T* pointer;
  typedef const T* const_pointer;

  template<typename _Tp1>
  struct rebind {
    typedef count_allocator<_Tp1> other;
  };

  pointer allocate(size_type n, const void *hint=0) {
#ifdef DEBUG
    fprintf(stderr, "Alloc %lu bytes.\n", n*sizeof(T));
#endif
    long long id = System::thread().getThreadSelf();
#ifndef SINGLE_THREAD
    {
      LocalSynchronizer local(lock);
#endif
      memuse[id] += n*sizeof(T);
#ifndef SINGLE_THREAD
    }
#endif
    return std::allocator<T>::allocate(n, hint);
  }

  void deallocate(pointer p, size_type n) {
#ifdef DEBUG
    fprintf(stderr, "Dealloc %lu bytes (%p).\n", n*sizeof(T), p);
#endif
    long long id = System::thread().getThreadSelf();
#ifndef SINGLE_THREAD
    {
      LocalSynchronizer local(lock);
#endif
      memuse[id] -= n*sizeof(T);
#ifndef SINGLE_THREAD
    }
#endif
    return std::allocator<T>::deallocate(p, n);
  }

  count_allocator() throw(): std::allocator<T>() { 
#ifdef DEBUG
    fprintf(stderr, "Hello allocator!\n"); 
#endif
  }
  count_allocator(const count_allocator &a) throw(): std::allocator<T>(a) { }
  template <typename U>                    
  count_allocator(const count_allocator<U> &a) throw(): std::allocator<T>(a) { }
  ~count_allocator() throw() { }
};



// Data structure to hold the dp row for one k-mer (both forward and reverse)
class map_element {
private:
  // dp row, only nonzero entries stored
  vector<pair<int, int>, count_allocator<pair<int, int> > > s_f;
  vector<pair<int, int>, count_allocator<pair<int, int> > > s_r;

#ifdef STATS
  int nonzeros;
#endif

public:
  // Construct new dp row with dd+1 elements
  map_element(int dd) {
#ifdef STATS
    nonzeros=0;
#endif
  }

  // Construct new dp row
  map_element() {
#ifdef STATS
    nonzeros=0;
#endif
  }

  // Get the i'th value of the dp row for forward k-mer
  // Most accesses are to the last (or second last) item so check those first
  // If accessing other items, use binary search
  int getf(int i) {
    if (s_f.size() == 0)
      return 0;

    if (s_f[s_f.size()-1].first == i)
      return s_f[s_f.size()-1].second;

    if (s_f.size() > 1 && s_f[s_f.size()-2].first == i)
      return s_f[s_f.size()-2].second;

#ifdef DEBUG
    for(int j = 0; j < s_f.size(); j++) {
      std::cout << "(" << s_f[j].first << "," << s_f[j].second << ") ";
    }
    std::cout << std::endl;

    std::cout << i << std::endl;
#endif

    // Binary search
    int lo = 0;
    int hi = s_f.size()-1;
    int mid = (lo + hi)/2;
    while(lo <= hi) {
#ifdef DEBUG
      std::cout << mid << " ";
#endif
      if (s_f[mid].first < i) {
	lo = mid+1;
      } else if (s_f[mid].first > i) {
	hi = mid-1;
      } else {
	return s_f[mid].second;
      }
      mid = (lo+hi)/2;
    }
#ifdef DEBUG
    std::cout << std::endl;
#endif
    return 0;
  }

  // Get the i'th value of the dp row for reverse k-mer
  // Most accesses are to the last (or second last) item so check those first
  // If accessing other items, use binary search
  int getr(int i) {
    if (s_r.size() == 0)
      return 0;

    if (s_r[s_r.size()-1].first == i)
      return s_r[s_r.size()-1].second;

    if (s_r.size() > 1 && s_r[s_r.size()-2].first == i)
      return s_r[s_r.size()-2].second;

#ifdef DEBUG
    for(int j = 0; j < s_r.size(); j++) {
      std::cout << "(" << s_r[j].first << "," << s_r[j].second << ") ";
    }
    std::cout << std::endl;

    std::cout << i << std::endl;
#endif

    // Binary search
    int lo = 0;
    int hi = s_r.size()-1;
    int mid = (lo + hi)/2;
    while(lo <= hi) {
#ifdef DEBUG
      std::cout << mid << " ";
#endif
      if (s_r[mid].first < i) {
	lo = mid+1;
      } else if (s_r[mid].first > i) {
	hi = mid-1;
      } else {
	return s_r[mid].second;
      }
      mid = (lo+hi)/2;
    }
#ifdef DEBUG
    std::cout << std::endl;
#endif
    return 0;
  }

  // Supports only updating the last value and adding a new value
  // (with larger index i than previous entries)
  void setf(int i, int v) {
    if (s_f.size() == 0) {
      s_f.push_back(std::make_pair(i,v));
#ifdef STATS
      nonzeros++;
#endif
      return;
    }

    if (s_f[s_f.size()-1].first == i) {
      s_f[s_f.size()-1].second = v;
    } else {
      if (s_f[s_f.size()-1].first > i) {
	// Just checking for misuse
	std::cout << "Setf will fail: " << i << " < " << s_f[s_f.size()-1].first << std::endl;
      }
      s_f.push_back(std::make_pair(i,v));
#ifdef STATS
      nonzeros++;
#endif
    }    
  }

  // Supports only updating the last value and adding a new value
  // (with larger index i than previous entries)
  void setr(int i, int v) {
    if (s_r.size() == 0) {
      s_r.push_back(std::make_pair(i,v));
#ifdef STATS
      nonzeros++;
#endif
      return;
    }

    if (s_r[s_r.size()-1].first == i) {
      s_r[s_r.size()-1].second = v;
    } else {
      if (s_r[s_r.size()-1].first > i) {
	// Just checking for misuse
	std::cout << "Setr will fail: " << i << " < " << s_r[s_r.size()-1].first << std::endl;
      }
      s_r.push_back(std::make_pair(i,v));
#ifdef STATS
      nonzeros++;
#endif
    }    
  }

#ifdef STATS
  int get_nz() {
    return nonzeros;
  }
#endif
};

// Providing hashes for Nodes
struct node_hash {
  size_t operator()(const Node & node) const{
    return oahash(node.kmer);
  }
};

// Fill a gap
// kmer_left: sequence of length k+fuz for extracting k-mers from the left end of the gap
// kmer_right: sequence of length k+fuz for extracting k-mers from the right end of the gap
// c and d: The length of the gap can vary between c and d
// fuz: maximum number of gap bordering nucleotides that can be ignored
// k: k-mer size
// The following should be non-null if a (ad hoc random) filling sequence should be recovered by traceback. 
// Otherwise they should all be NULL.
// left_fuz: The number of actual ignored nucleotides on the left end of the gap
// right_fuz: The number of actual ignored nucleotides on the right end of the gap
// fill: The sequence to fill the gap. Start offset of actual sequence in fill is fuz - left_fuz, 
//       the fill seq always includes the right kmer
int Gap2Seq::fill_gap(Graph graph, std::string kmer_left, std::string kmer_right, int gap_len, int gap_err, int fuz, int k, long long max_mem, int *left_fuz, int *right_fuz, char *fill) {

  long long id = System::thread().getThreadSelf();
#ifndef SINGLE_THREAD
  {
      LocalSynchronizer local(lock);
#endif
      memuse[id] = 0;
#ifndef SINGLE_THREAD
  }
#endif
  // Border sets for breadth first search
  std::unordered_set<Node, node_hash, equal_to<Node>, count_allocator<Node> > border;
  std::unordered_set<Node, node_hash, equal_to<Node>, count_allocator<Node> > nextBorder;

  // The number of paths
  int count = 0;

  // The set of nodes reached so far
  unordered_map<Node, map_element *, node_hash, equal_to<Node>, count_allocator< pair <const Node, map_element *> > > reachableSet;

#ifdef DEBUG
  std::cout << "Kmer " << kmer_left << std::endl;
#endif

  // Leftmost starting k-mer
  string kmer = kmer_left.substr(0,k);
  Node node = graph.buildNode(Data((char *)kmer.c_str()));

#ifdef DEBUG
  if (!graph.contains(node)) {
    std::cout << "Kmer " << kmer << " cannot be found in the graph!" << std::endl;
  }
#endif

  // Put the leftmost k-mer into border and reachable set
  int currentD = 0;

  if (graph.contains(node)) {
    border.insert(node);
    map_element *me = NULL;

    if (reachableSet.find(node) != reachableSet.end()) {
      me = reachableSet[node];
    }
      
    if (me == NULL) {
      me = new map_element(gap_len+gap_err+2*fuz);
      reachableSet[node] = me;
    }
    
    // Initialize the dp row
    if (node.strand == STRAND_FORWARD) {
      me->setf(currentD, 1);
    } else {
      me->setr(currentD, 1);
    }
  }


  currentD++;

  // BFS loop until depth d+2*fuz is reached
  long long mymemuse;
#ifndef SINGLE_THREAD
  {
      LocalSynchronizer local(lock);
#endif
      mymemuse = memuse[id];
#ifndef SINGLE_THREAD
  }
#endif

  while(currentD <= gap_len+gap_err+2*fuz && mymemuse < max_mem && count == 0) {
#ifdef DEBUG
    std::cout << "Memuse: " << mymemuse << std::endl;
#endif

    // Iterate over the nodes at current depth
#ifdef DEBUG
    std::cout << "Border size: " << border.size() << std::endl;
#endif

    for (std::unordered_set<Node, node_hash, equal_to<Node>, count_allocator<Node> >::iterator it = border.begin(); it != border.end(); ++ it) {
#ifndef SINGLE_THREAD
      {
	LocalSynchronizer local(lock);
#endif
	mymemuse = memuse[id];
#ifndef SINGLE_THREAD
      }
#endif

      if (mymemuse >= max_mem)
	break;

      Node n = (Node) *it;
      
#ifdef DEBUG
      std::cout << graph.toString(n) << (n.strand == STRAND_FORWARD ? " F": " R") <<  std::endl;
#endif
      
      // Get the neightbors of the node and update their dp rows
      Graph::Vector<Node> neighbors = graph.successors<Node>(n);
      map_element *node_me = reachableSet[n];
      // Number of paths to the parent node
      int num_paths = n.strand == STRAND_FORWARD ? node_me->getf(currentD-1) : node_me->getr(currentD-1);

      for (int i = 0; i < neighbors.size(); i++) {
	map_element *me = NULL;

	if (reachableSet.find(neighbors[i]) != reachableSet.end()) {
	  me = reachableSet[neighbors[i]];
	}

	if (me == NULL) {
	  me = new map_element(gap_len+gap_err+2*fuz);
	  reachableSet[neighbors[i]] = me;
	}
	if (neighbors[i].strand == STRAND_FORWARD) {
	  me->setf(currentD, me->getf(currentD) + num_paths > MAX_PATHS ? MAX_PATHS : me->getf(currentD) + num_paths);
	} else {
	  me->setr(currentD, me->getr(currentD) + num_paths > MAX_PATHS ? MAX_PATHS : me->getr(currentD) + num_paths);
	}

#ifdef DEBUG
	std::cout << "Neighbor: " << graph.toString(neighbors[i]) << (neighbors[i].strand == STRAND_FORWARD ? " F": " R") <<  std::endl;
#endif
	// Put the neighbor to the next border
	nextBorder.insert(neighbors[i]);
      }
    }
    
#ifndef SINGLE_THREAD
    {
      LocalSynchronizer local(lock);
#endif
      mymemuse = memuse[id];
#ifndef SINGLE_THREAD
    }
#endif
    if (mymemuse >= max_mem)
      break;

    // Swap the borders
    border.clear();
    border.swap(nextBorder);

    // Add next starting k-mer if one still exists
    if (currentD <= fuz) {
      string kmer = kmer_left.substr(currentD, k);
      Node node = graph.buildNode(Data((char *)kmer.c_str()));
#ifdef DEBUG
      if (!graph.contains(node)) {
	std::cout << "Kmer " << kmer_left << " cannot be found in the graph!" << std::endl;
      }
#endif

#ifdef DEBUG
      std::cout << "Kmer: " << kmer << std::endl;
#endif

      if (graph.contains(node)) {
	border.insert(node);
	map_element *me = NULL;

	if (reachableSet.find(node) != reachableSet.end()) {
	  me = reachableSet[node];
	}

	if (me == NULL) {
	  me = new map_element(gap_len+gap_err+2*fuz);
	  reachableSet[node] = me;
	}

	// Initialize the dp row. Note that here we initialize to 1
	// regardless of whether the node has already been reached. This
	// will prevent counting paths going via all the starting k-mers
	// several times
	if (node.strand == STRAND_FORWARD) {
	  me->setf(currentD, 1);
	} else {
	  me->setr(currentD, 1);
	}
      }
    }

    // Check if a path has been found
    if (currentD >= gap_len+2*fuz) {
      int err = currentD - gap_len - 2*fuz;

      // Iterate over the k-mers in the right edge of the gap
      for (int j = 0; j <= fuz && count == 0; j++) {
	string rkmer = kmer_right.substr(j,k);
	Node rnode = graph.buildNode(Data((char *)rkmer.c_str()));

#ifdef DEBUG
	std::cout << "Right k-mer: " << graph.toString(rnode) << std::endl;
	if (!graph.contains(rnode)) {
	  std::cout << "Kmer " << kmer_right << " cannot be found in the graph!" << std::endl;
	}
#endif

	map_element *right = NULL;

	if (reachableSet.find(rnode) != reachableSet.end())
	  right = reachableSet[rnode];
	std::vector<int> pathLengths;

	// Check if this k-mer has been reached
	if (right == NULL) {
#ifdef DEBUG
	  std::cout << "right kmer not reached" << std::endl;
#endif
	} else {
	  // +fuz: path lengths are counted starting from the first allowed left k-mer
	  // +j: j nucleotides were ignored on the right side thus widening the gap
	  int actual_gap_len1 = gap_len + fuz + j + err;
	  int actual_gap_len2 = gap_len + fuz + j - err;

	  if (rnode.strand == STRAND_FORWARD) {
	    if (right->getf(actual_gap_len1) >= 1) {
	      count = count + right->getf(actual_gap_len1) > MAX_PATHS ? MAX_PATHS : count + right->getf(actual_gap_len1);
#ifdef DEBUG
	      std::cout << "Gap filled: " << actual_gap_len1 << "\t" << right->getf(actual_gap_len1) << std::endl;
#endif
	      pathLengths.push_back(actual_gap_len1);
	    }
	  } else {
	    if (right->getr(actual_gap_len1) >= 1) {
	      count = count + right->getr(actual_gap_len1) > MAX_PATHS ? MAX_PATHS : count + right->getr(actual_gap_len1);
#ifdef DEBUG
	      std::cout << "Gap filled: " << actual_gap_len1 << "\t" << right->getr(actual_gap_len1) << std::endl;
#endif
	      pathLengths.push_back(actual_gap_len1);
	    }
	  }
	
	  if (actual_gap_len2 != actual_gap_len1 && actual_gap_len2 >= 0) {
	    if (rnode.strand == STRAND_FORWARD) {
	      if (right->getf(actual_gap_len2) >= 1) {
		count = count + right->getf(actual_gap_len2) > MAX_PATHS ? MAX_PATHS : count + right->getf(actual_gap_len2);
#ifdef DEBUG
		std::cout << "Gap filled: " << actual_gap_len2 << "\t" << right->getf(actual_gap_len2) << std::endl;
#endif
		pathLengths.push_back(actual_gap_len2);
	      }
	    } else {
	      if (right->getr(actual_gap_len2) >= 1) {
		count = count + right->getr(actual_gap_len2) > MAX_PATHS ? MAX_PATHS : count + right->getr(actual_gap_len2);
#ifdef DEBUG
		std::cout << "Gap filled: " << actual_gap_len2 << "\t" << right->getr(actual_gap_len2) << std::endl;
#endif
		pathLengths.push_back(actual_gap_len2);
	      }
	    }
	  }
	}

	if (count > 0 && pathLengths.size() > 0) {
	  // Choose by random the length of the path to follow
	  int currentD2 = pathLengths[rand() % pathLengths.size()];
	  if (fill != NULL) {
	    // Recover the fill sequence
	
	    // Number of ignored nucleotides on the right edge
	    *right_fuz = j;
	
	    // Trace back in the dp matrix
	  
	    Node current = rnode;
	    // Set of in neighbors of the current node
	    std::vector<Node> back;
	    fill[currentD2] = '\0';
	  
	    while(currentD2 >= 0) {
	      // The current k-mer
	      string str = graph.toString(current);
	  
#ifdef DEBUG
	      std::cout << str << std::endl;
#endif
	  
	      // Check for end condition
	      if (currentD2 <= fuz) {
		Node lnode = graph.buildNode(Data((char *)kmer_left.substr(currentD2,k).c_str()));
#ifdef DEBUG
		std::cout << kmer_left << " " << graph.toString(lnode) << " " << graph.toString(current) << std::endl;
#endif
		if (lnode == current) {
		  *left_fuz = fuz-currentD2;
		  break;
		}
	      }

	      // Get the in neighbors
	      if (currentD2 > 0) {
		fill[currentD2-1] = str[str.length()-1];
		Graph::Vector<Node> neighbors = graph.predecessors<Node>(current);
		for (int i = 0; i < neighbors.size(); i++) {
		  Node n = neighbors[i];
		  if (reachableSet.find(n) != reachableSet.end()) {
		    if (n.strand == STRAND_FORWARD) {
		      if (reachableSet[n]->getf(currentD2-1) > 0) {
			back.push_back(n);
		      }
		    } else {
		      if (reachableSet[n]->getr(currentD2-1) > 0) {
			back.push_back(n);
		      }
		    }
		  }
		}
		// There should be in-neighbors where the path originated from but check just in case...
		if (back.size() == 0) {
		  std::cout << "Unable to backtrace! " << currentD2 << " " << currentD << " " << graph.toString(rnode) <<  std::endl;
		  // Free memory
		  for(auto it = reachableSet.begin(); it != reachableSet.end(); ++it) {
		    delete it->second;
		  }
		  return 0;
		}
		// Randomly choose one of the in neighbors
		current = back[rand()%(back.size())];
	      }
	      currentD2--;
	      back.clear();
	    }
	  }
	  break;
	}
      }
    }

    currentD++;
#ifndef SINGLE_THREAD
    {
      LocalSynchronizer local(lock);
#endif
      mymemuse = memuse[id];
#ifndef SINGLE_THREAD
    }
#endif
  }
#ifdef DEBUG
  std::cout << std::endl;
#endif

#ifdef DEBUG
  std::cout << "Reachable set: " << std::endl;
  for(auto it = reachableSet.begin(); it != reachableSet.end(); ++it) {
    std::cout << graph.toString(it->first) << std::endl;
    for(int i = 0; i <= gap_len+gap_err+2*fuz; i++) {
      if (it->first.strand == STRAND_FORWARD) {
	std::cout << " " << (it->second)->getf(i);
      } else {
	std::cout << " " << (it->second)->getr(i);
      }
    }
    std::cout << std::endl;
  }
#endif

  // Free memory
#ifdef STATS
  long long nz = 0;
  long long n = 0;
#endif
  for(auto it = reachableSet.begin(); it != reachableSet.end(); ++it) {
#ifdef STATS
    n++;
    nz += it->second->get_nz();
#endif
    delete it->second;
  }

#ifdef STATS
#ifndef SINGLE_THREAD
  {
      LocalSynchronizer local(lock);
#endif
      std::cout << "Nodes: " << n << " Nonzeros: " << nz << " Total memory: " << memuse[id] << std::endl; 
#ifndef SINGLE_THREAD
  }
#endif
#endif

  return count;
}

std::set<Node> Gap2Seq::extract_reachable_nodes(Graph graph, std::string kmer, int d) {
  std::set<Node> border;
  std::set<Node> nextBorder;
  int currentD = 0;

  std::set<Node> reachableSet;

  Node node = graph.buildNode(Data((char *)kmer.c_str()));

  if (!graph.contains(node)) {
    std::cout << "Kmer " << kmer << " cannot be found in the graph!" << std::endl;
    exit(EXIT_FAILURE);
  }

  border.insert(node);

  while(currentD <= d) {
    for (std::set<Node>::iterator it = border.begin(); it != border.end(); ++ it) {
      Node n = (Node) *it;
      reachableSet.insert(n);
      Graph::Vector<Node> neighbors = graph.successors<Node>(n);
      for (int i = 0; i < neighbors.size(); i++) {
	nextBorder.insert(neighbors[i]);
      }
    }

    border.clear();
    border.swap(nextBorder);
    currentD++;
  }

  return reachableSet;
}
